# Copyright 2019 IncQuery Labs Ltd.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#  http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from iqs_jupyter import api_composition
'''
Created on 2019-04-09

@author: Gábor Bergmann
'''

from typing import Optional
import html
import collections


import ipywidgets as widgets
from IPython.display import display

import iqs_client


## TODO separate TWC-specific, MMS-specific and core parts into separate files if possible (re-export of core?)
## TODO refactor: pull up TWC-independent parts of (a) revision selector widget (retrofit mms commit selector) and (b) OSMC element info widget

import iqs_jupyter.config_defaults as defaults
import iqs_jupyter.tool_extension_point as ext_point
from iqs_jupyter.helpers import cell_to_html as _cell_to_html
from iqs_jupyter.helpers import dict_to_element as _dict_to_element

# recognizing element-in-compartment descriptors in match results

def _recognize_element_in_compartment_descriptor(
        dict_of_element : dict
) -> Optional[iqs_client.ElementInCompartmentDescriptor]:
    if "relativeElementID" in dict_of_element:
        return iqs_client.ElementInCompartmentDescriptor(
            **dict(
                (py_name, dict_of_element[json_name]) 
                for py_name, json_name in iqs_client.ElementInCompartmentDescriptor.attribute_map.items()
            )
        )
    else:
        return None
        
ext_point.element_dict_recognizers.append(_recognize_element_in_compartment_descriptor)


def _recognize_typed_element_in_compartment_descriptor(
        dict_of_element : dict
) -> Optional[iqs_client.TypedElementInCompartmentDescriptor]:
    if "element" in dict_of_element:
        return iqs_client.TypedElementInCompartmentDescriptor(
            **dict(
                (
                    py_name, 
                    _recognize_element_in_compartment_descriptor(dict_of_element[json_name]) 
                        if py_name == "element" else dict_of_element[json_name]
                ) 
                for py_name, json_name in iqs_client.TypedElementInCompartmentDescriptor.attribute_map.items()
            )
        )
    else:
        return None
        
ext_point.element_dict_recognizers.append(_recognize_typed_element_in_compartment_descriptor)




## monkey patch section

def _monkey_patch_element_in_compartment_descriptor_repr_html(self):
    if hasattr(self, 'url'):
        return '<a href="{}" title="{}">element #{}</a>'.format(html.escape(self.url), html.escape(self),
                                                                html.escape(self.relative_element_id))
    else:
        return '<span title="{}">element #{}</span>'.format(html.escape(self.to_str()),
                                                            html.escape(self.relative_element_id))



def _monkey_patch_element_in_compartment_descriptor_resolve_reference(self, target_element_relative_id):
    return iqs_client.ElementInCompartmentDescriptor(
        compartment_uri     = self.compartment_uri,
        relative_element_id = target_element_relative_id
    )

def _monkey_patch_model_compartment_get_element_in_compartment_by_id(self, relative_element_id):
    return iqs_client.ElementInCompartmentDescriptor(
        compartment_uri=self.compartment_uri,
        relative_element_id=relative_element_id
    )

def _monkey_patch_model_compartment_is_loaded_by_server(self, iqs):
    in_memory_compartments_response = iqs.in_memory_index.list_inmemory_model_compartments()
    return self in in_memory_compartments_response.inmemory_model_compartments


def _monkey_patch_query_execution_response_to_data_frame(self, url_provider=None):
    import pandas as pd
    return pd.DataFrame(self.to_list_of_matches(url_provider))

def _monkey_patch_query_execution_response_to_list_of_matches(self, url_provider=None):
    return [dict((arg.parameter, _dict_to_element(arg.value, url_provider)) for arg in match.arguments) for match in
            self.matches]

#****
    
def _monkey_patch_typed_element_in_compartment_repr_html(self):
    type_html = html.escape(self.type if self.type else "<untyped>")
    element_html = self.element._repr_html_()
    if self.name:
        return "'{}' ({}) {}".format(html.escape(self.name), type_html, element_html)
    elif self.type:
        return "<unnamed> ({}) {}".format(type_html, element_html)
    else:
        return element_html

def _monkey_patch_generic_validation_rule_repr_html(self):
    return '<span title="{}">{} occurrences of "{}" <i>(see hover for details)</i></span> {}'.format(
        html.escape(self.to_str()),
        len(self.matching_elements),
        html.escape(self.message),
        "<ul>{}</ul>".format(
            "\n".join([
                "<li>{}</li>".format(
                    typed_diag_element.element._repr_html_()
                )
                for typed_diag_element in self.matching_elements
            ])
        )
    )
    
    
def _monkey_patch_generic_validation_results_to_data_frame(self):
    import pandas as pd
    return pd.DataFrame([item._asdict() for item in self.to_list_of_diagnostic_items()])


def _monkey_patch_generic_validation_results_to_list_of_diagnostic_items(self):
    return [
        ValidationDiagnosticItem(
            rule.constraint.type, rule.constraint.name, rule.constraint.element.relative_element_id, rule.constraint.element,
            rule.severity, validation_marker.message,
            validation_marker.matching_element.type, validation_marker.matching_element.name, 
            validation_marker.matching_element.element.relative_element_id, validation_marker.matching_element.element
        )
        for rule in self.rules
        for validation_marker in rule.matching_elements
    ]
    
def _monkey_patch_generic_validation_results_repr_html(self):
    diag_entries = self.to_list_of_diagnostic_items()
    table_html = "" if not diag_entries else '''
        <div>
            <table border="1">
                <thead>
                    <tr style="text-align: right;">
                        <th>Severity</th>
                        <th>Rule name</th>
                        <th>Message</th>
                        <th>Element type</th>
                        <th>Element name</th>
                        <th>Element</th>
                    </tr>
                </thead>
                <tbody>
                    {}
                </tbody>
            </table>
        </div>
    '''.format(
        "\n".join([
            "<tr>{}</tr>".format(
                " ".join([
                    "<td>{}</td>".format(field)
                    for field in [
                        html.escape(diag_entry.severity),
                        html.escape(diag_entry.constraint_element_name), 
                        html.escape(diag_entry.message), 
                        html.escape(diag_entry.matching_element_type), 
                        html.escape(diag_entry.matching_element_name), 
                        diag_entry.matching_element._repr_html_()
                    ]
                ])
            )
            for diag_entry in diag_entries
        ])
    )
    return '<span title="{}">Validation results: {} diagnostic items <i>(see hover for details)</i></span> {}'.format(
        html.escape(self.to_str()),
        len(diag_entries),
        table_html
    )
    
def _monkey_patch_validation_diagnostics_reps_html(self):
    return '<span title="{}">Validation results summary <i>(see hover for details)</i></span> <ul>{}</ul>'.format(
        html.escape(self.to_str()),
        "\n".join([
            "<li>{} ({}) {}</li>".format(
                html.escape(diag_type), 
                str(diag_count),
                ""
#                 "" if not diag_count else "<ul>{}</ul>".format(
#                     "\n".join([
#                         '<li>{}</li>'.format(
#                             rule._repr_html_()
#                         )
#                         for rule in self.rules
#                         if rule.severity == diag_type and rule.matching_elements
#                     ])
#                 )
            )
            for diag_type, diag_count in self.to_dict().items()
        ])
    )
    
def _monkey_patch_generic_response_message_repr_html_(self):
    return '<span title="{}">{} <i>(see hover for details)</i></span>'.format(html.escape(self.to_str()), html.escape(self.message))

def _monkey_patch_query_fqn_list_repr_html_(self):
    ns_list = self.query_fq_ns
    if ns_list:
        list_body = "\n".join([
            "<li>{}</li>".format(html.escape(query_fqn)) for query_fqn in ns_list
        ])
        return '<span title="{}">Listing {} queri(es): <i>(see hover for details)</i></span> <ul>{}</ul>'.format(
            html.escape(self.to_str()), str(len(ns_list)), list_body)
    return '<span title="{}">No queries <i>(see hover for details)</i></span>'.format(html.escape(self.to_str()))

def _monkey_patch_query_execution_response_to_html(self):
    count_report = '<span title="{}">{} match(es) of query{}{} <i>(see hover for details)</i></span>'.format(
        html.escape(self.to_str()), 
        html.escape(str(self.match_set_size)),
        " '{}'".format(html.escape(self.query_fqn)) if self.query_fqn else "",
        " with bound parameter(s) '{}'".format(", ".join([html.escape(binding.parameter) for binding in self.binding])) if self.binding else ""
    )
    match_list = self.to_list_of_matches()
    
    if not match_list:
        return count_report
    
    pattern_params = match_list[0].keys()
    
    param_headers = " ".join(["<th>{}</th> ".format(html.escape(param)) for param in pattern_params])
    match_rows = "\n".join([
        "<tr><th>{}</th>{}</tr>\n".format(row_index," ".join([
            " <td>{}</td>".format(_cell_to_html(match_list[row_index][param])) for param in pattern_params
        ])) for row_index in range(len(match_list))
    ])
    
    param_bindings = " ".join([
        "<td>{}</td> ".format(" ".join([ # 0 or 1
            _cell_to_html(_dict_to_element(binding.value))
            for binding in self.binding
            if binding.parameter == param
        ]))
        for param in pattern_params
    ])
    param_bindings_row = '''
        <tr>
            <th></th>
            {}
        </tr>    
    '''.format(param_bindings)
    
    style = ""
    return '''
        {}
        <div>
        {}
        <table border="1">
            <thead>
                <tr style="text-align: right;">
                    <th></th>
                    {}
                </tr>
                {}
            </thead>
            <tbody>
                {}
            </tbody>
        </table>
        </div>
    '''.format(count_report, style, param_headers, param_bindings_row, match_rows)

        

def _do_monkey_patching():
    iqs_client.ElementInCompartmentDescriptor._repr_html_ = _monkey_patch_element_in_compartment_descriptor_repr_html
    iqs_client.ElementInCompartmentDescriptor.resolve_reference = _monkey_patch_element_in_compartment_descriptor_resolve_reference

    iqs_client.ModelCompartment.get_element_in_compartment_by_id = _monkey_patch_model_compartment_get_element_in_compartment_by_id
    iqs_client.ModelCompartment.is_loaded_by_server = _monkey_patch_model_compartment_is_loaded_by_server
    
    iqs_client.QueryExecutionResponse.to_list_of_matches = _monkey_patch_query_execution_response_to_list_of_matches
    iqs_client.QueryExecutionResponse.to_data_frame = _monkey_patch_query_execution_response_to_data_frame
    iqs_client.QueryExecutionResponse.to_data_frame = _monkey_patch_query_execution_response_to_data_frame
    iqs_client.QueryExecutionResponse._repr_html_ = _monkey_patch_query_execution_response_to_html  
    iqs_client.IndexMessage._repr_html_ = _monkey_patch_generic_response_message_repr_html_
    iqs_client.SimpleMessage._repr_html_ = _monkey_patch_generic_response_message_repr_html_
    iqs_client.QueryFQNList._repr_html_ = _monkey_patch_query_fqn_list_repr_html_
    
    # TODO TWC-specific version missing, as well as additional features
    iqs_client.GenericValidationResults.to_list_of_diagnostic_items = _monkey_patch_generic_validation_results_to_list_of_diagnostic_items
    iqs_client.GenericValidationResults.to_data_frame = _monkey_patch_generic_validation_results_to_data_frame
    iqs_client.GenericValidationResults._repr_html_ = _monkey_patch_generic_validation_results_repr_html
    iqs_client.GenericValidationRule._repr_html_ = _monkey_patch_generic_validation_rule_repr_html
    iqs_client.ValidationDiagnostics._repr_html_ = _monkey_patch_validation_diagnostics_reps_html
    iqs_client.TypedElementInCompartmentDescriptor._repr_html_ = _monkey_patch_typed_element_in_compartment_repr_html

_do_monkey_patching()


## global utility functions

validation_color_scale = {
    'fatal': 'black', 
    'error': 'red', 
    'warning': 'orange', 
    'info': 'blue', 
    'debug': 'green'
}
validation_diagnostics_ranking = [
    'fatal', 
    'error', 
    'warning', 
    'info', 
    'debug'
]

def binding(**kwargs):
    return [
        {"parameter": param_name, "value": param_value} 
        for param_name, param_value in kwargs.items()
    ]

def connect(
    address  = defaults.default_IQS_address,
    user     = defaults.default_IQS_username,
    password = defaults.default_IQS_password,
    auth_with_user_pw  = True    
):
    configuration = iqs_client.Configuration()
    configuration.host=address
    configuration.access_token = None
    if auth_with_user_pw:
        configuration.username=user
        configuration.password=password
    return IQSClient(configuration)


## useful widgets and clients

class IQSConnectorWidget:
    def __init__(
        self,
        ask_for_user_pw  = True,
        initial_address  = defaults.default_IQS_address,
        initial_user     = defaults.default_IQS_username,
        initial_password = defaults.default_IQS_password,
        address_label    = 'Address:',
        user_label       = 'User:',
        password_label   = 'Password:',
        labelText="IQS API Access Point",
        auto_display=True
    ):
        self.ask_for_user_pw = ask_for_user_pw
        
        self.address_field  = widgets.Text    (value=initial_address,  description=address_label)
        self.user_field     = widgets.Text    (value=initial_user,     description=user_label)
        self.password_field = widgets.Password(value=initial_password, description=password_label)
        
        if ask_for_user_pw:
            fields = [self.address_field, self.user_field, self.password_field]
        else:
            fields = [self.address_field]
        
        self.box=widgets.HBox([widgets.Label(value=labelText), widgets.VBox(fields)])
        if auto_display: self.display()
            
    def display(self):
        display(self.box)
        
    def _repr_html_(self):
        self.display()
        
    def api_client_configuration(self):
        configuration = iqs_client.Configuration()
        configuration.host=self.address_field.value 
        if self.ask_for_user_pw:
            configuration.username=self.user_field.value
            configuration.password=self.password_field.value
        return configuration
    
    def connect(self):
        return connect(
            address           = self.address_field.value,
            user              = self.user_field.value,
            password          = self.password_field.value,
            auth_with_user_pw = self.ask_for_user_pw
        )






class IQSClient:
    def __init__(
        self,
        root_configuration
    ):
        api_composition.decorate_iqs_client(self, root_configuration)
        self.jupyter_tools = ext_point.IQSJupyterTools(self)



# custom validation result datatype 

ValidationDiagnosticItem = collections.namedtuple(
    "ValidationDiagnosticItem", 
    ['constraint_element_type', 'constraint_element_name', 'constraint_element_relative_id', 'constraint_element',
     'severity', 'message', 
     'matching_element_type', 'matching_element_name', 'matching_element_relative_id', 'matching_element']
)




# Copyright 2019 IncQuery Labs Ltd.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#  http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Created on 2019. máj. 24.

@author: Gabor Bergmann
'''

# single-sourcing of version string
# see solution 3 @ https://packaging.python.org/guides/single-sourcing-package-version
__version__ = "0.20.0.dev2"



# Copyright 2019 IncQuery Labs Ltd.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#  http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ipywidgets.widgets.widget import Widget

'''
Created on 2019. júl. 4.

@author: Gabor Bergmann
'''

from abc import ABC, abstractmethod
from typing import Dict, Tuple, Any, Hashable, List

import ipywidgets as widgets
from IPython.display import display

class AbstractElementInfoWidget(ABC):
    def __init__(
        self,
        #initial_element,
        attempt_batch_prefetch = True,
        grid_container_layout = widgets.Layout(
            grid_template_columns = "25% 73%", 
            grid_gap="2%"),
        field_box_layout = widgets.Layout(
            overflow_x='auto', 
            justify_content='flex-end'),
        slot_item_layout = widgets.Layout(
            width='auto', 
            flex='0 0 auto'),
        slot_box_layout = widgets.Layout(
            overflow_x='auto',
            #border='solid',
            width='auto',
            height='',
            flex_flow='row',
            display='flex', 
            justify_content='flex-start')
    ):
        self.attempt_batch_prefetch = attempt_batch_prefetch
        self.grid_container_layout = grid_container_layout
        self.field_box_layout = field_box_layout
        self.slot_item_layout = slot_item_layout
        self.slot_box_layout = slot_box_layout
        self.accordion = widgets.Accordion()
        self.element_list = []
        self.element_id_to_index = {}
        #self.initial_element = initial_element
        #self.show_element_pane(initial_element)
     
    @abstractmethod
    def _to_hashable_identifier(self, element) -> Hashable:
        pass
    
    @abstractmethod
    def _to_field_display_name(self, element, field) -> str:
        pass
    
    @abstractmethod
    def _unpack_element(self, element) -> Tuple[str, Dict[str, Any]]:
        pass
    
    @abstractmethod
    def _do_retrieve_batch(self, element, field_values):
        pass
    
    @abstractmethod
    def _value_widget_list(self, element, field, value) -> List[Widget]:
        pass
        
    def _add_element_pane(self, element):
        header, field_values = self._unpack_element(element)

        if self.attempt_batch_prefetch:
            self._do_retrieve_batch(element, field_values)
            
        import html
        grid_content_boxes = []
        for field,value in field_values.items():
            field_display_name = self._to_field_display_name(element, field)
            if field_display_name and (value is not None) and not ([]==value):
                field_widget = widgets.HTML(value = "<b>{}</b>".format(html.escape(field_display_name)))
                field_box = widgets.Box(children=[field_widget], layout=self.field_box_layout)
                slot_widget_list = self._value_widget_list(element, field, value)
                slot_box = widgets.Box(children=slot_widget_list, layout=self.slot_box_layout)
                grid_content_boxes.extend([field_box, slot_box])
        content = widgets.GridBox(layout=self.grid_container_layout, children=grid_content_boxes)
        
        self.accordion.children = list(self.accordion.children) + [content]
        self.accordion.set_title(len(self.element_list), header)
        self.element_id_to_index[self._to_hashable_identifier(element)] = len(self.element_list)
        self.element_list.append(element)


    def _create_value_widget_attribute(self, value):
        # TODO handle data types separately, e.g. checkbox for boolean
        value_str = value if isinstance(value, str) else str(value)
        return widgets.Label(value=value_str, layout=self.slot_item_layout)
    def _create_value_widget_reference(self, referred_element, short_text, long_text):
        button = widgets.Button(
            description=short_text,
            #disabled=False,
            button_style='info', # 'primary', 'success', 'info', 'warning', 'danger' or ''
            #icon='check',
            tooltip=long_text,
            layout=self.slot_item_layout
        )
        button.on_click(lambda b: self.show_element_pane(referred_element))
        return button


    def show_element_pane(self, element, highlight=True):
        element_id = self._to_hashable_identifier(element)
        if element_id not in self.element_id_to_index:
            self._add_element_pane(element)
        if highlight:
            self.accordion.selected_index = self.element_id_to_index[element_id]
        
    def selected_element(self):
        if self.accordion.selected_index is None:
            return None
        else:
            return self.element_list[self.accordion.selected_index]

    def _repr_html_(self):
        self.display()    
          
    def display(self):
        display(self.accordion)



# Copyright 2019 IncQuery Labs Ltd.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#  http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ipywidgets.widgets.widget import Widget
from builtins import super

'''
Created on 2019. máj. 24.

@author: Gabor Bergmann
'''

from typing import Dict, Tuple, Any, Hashable, List

import ipywidgets as widgets
from IPython.display import display


import iqs_jupyter.config_defaults as defaults
import iqs_jupyter.abstract_widgets as abstract_widgets


class OSMCConnectorWidget:
    def __init__(
        self,
        initial_address = defaults.default_twc_osmc_address,
        initial_user    = defaults.default_twc_osmc_username,
        initial_pw      = defaults.default_twc_osmc_password,
        address_label='Address:',
        user_label='User:',
        pw_label='Password:',
        label_text="OSMC API Access Point",
        auto_display=True
    ):
        self.address_widget = widgets.Text    (value=initial_address, description=address_label)
        self.user_widget    = widgets.Text    (value=initial_user,    description=user_label)
        self.pw_widget      = widgets.Password(value=initial_pw,      description=pw_label)
        self.box=widgets.HBox([
            widgets.VBox([widgets.Label(value=label_text), widgets.Label(value=""), widgets.Label(value="")]),
            widgets.VBox([self.address_widget,             self.user_widget,        self.pw_widget])
        ])
        if auto_display: self.display()
            
    def display(self):
        display(self.box)
        
    def _repr_html_(self):
        self.display()    
          
    def connect(self, cache_locally = True):
        return OSMCClient(self.address_widget.value, (self.user_widget.value, self.pw_widget.value), cache_locally)


class OSMCClient:
    def __init__(
        self,
        access_point_prefix,
        auth,
        cache_locally = True
    ):
        if (access_point_prefix.endswith('/')):
            self.access_point_prefix = access_point_prefix
        else:
            self.access_point_prefix = access_point_prefix + '/'
        self.auth = auth
        self.cache_locally = cache_locally
        
    def element_api_link(self, element_descriptor):
        return "{}workspaces/{}/resources/{}/branches/{}/revisions/{}/elements/{}".format(
            self.access_point_prefix,
            element_descriptor.workspace_id,
            element_descriptor.resource_id,
            element_descriptor.branch_id,
            element_descriptor.revision_number,        
            element_descriptor.element_id
        )
        
    def element_batch_api_link(self, revision_descriptor):
        return "{}workspaces/{}/resources/{}/branches/{}/revisions/{}/elements".format(
            self.access_point_prefix,
            revision_descriptor.workspace_id,
            revision_descriptor.resource_id,
            revision_descriptor.branch_id,
            revision_descriptor.revision_number
        )
        
    def as_url_provider(self):
        def url_provider(element_descriptor):
            return self.element_api_link(element_descriptor)
        return url_provider
        
    def retrieve(self, element_descriptor):
        if hasattr(element_descriptor, 'raw_container_content') and hasattr(element_descriptor, 'raw_element_content'):
            return element_descriptor.raw_container_content, element_descriptor.raw_element_content
        else:
            return self.force_retrieve(element_descriptor)

    def force_retrieve(self, element_descriptor):
        raw_container_content, raw_element_content = self._retrieve_single(element_descriptor)
        if self.cache_locally:
            element_descriptor.raw_container_content = raw_container_content
            element_descriptor.raw_element_content = raw_element_content
        return raw_container_content, raw_element_content

    def _retrieve_single(self, element_descriptor):
        import requests
        req_url = self.element_api_link(element_descriptor)

        session = requests.session()
        osmc_response = session.get(req_url, auth=self.auth)
        osmc_response.raise_for_status()
        logout_response = session.get(f'{self.access_point_prefix}logout')
        logout_response.raise_for_status()

        raw_container_content, raw_element_content = osmc_response.json()
        return raw_container_content, raw_element_content
        
    def retrieve_batch(self, element_descriptor_list):
        if self.cache_locally:
            action_needed = [
                element_descriptor 
                for element_descriptor in element_descriptor_list 
                if not (hasattr(element_descriptor, 'raw_container_content') and hasattr(element_descriptor, 'raw_element_content'))
            ]
            if action_needed:
                self.force_retrieve_batch(action_needed)

    def force_retrieve_batch(self, element_descriptor_list):
        from itertools import groupby
        grouping = groupby(element_descriptor_list, lambda element_descriptor: element_descriptor.get_containing_revision())
        for revision_descriptor, group_with_same_descriptor in grouping:
            raw_content_dict = self._retrieve_batch(revision_descriptor, group_with_same_descriptor)
            if self.cache_locally:
                for element_descriptor in element_descriptor_list:
                    element_result_dict = raw_content_dict[element_descriptor.element_id]
                    if element_result_dict["status"] == 200: 
                        raw_container_content, raw_element_content = element_result_dict["data"]
                        element_descriptor.raw_container_content = raw_container_content
                        element_descriptor.raw_element_content = raw_element_content

    def _retrieve_batch(self, revision_descriptor, element_descriptor_list):
        element_id_set = set([element_descriptor.element_id for element_descriptor in element_descriptor_list])
        import requests
        req_url = self.element_batch_api_link(revision_descriptor)
        req_body = ','.join(element_id_set)

        session = requests.session()
        osmc_response = session.post(req_url, data=req_body, auth=self.auth)
        osmc_response.raise_for_status()
        logout_response = session.get(f'{self.access_point_prefix}logout')
        logout_response.raise_for_status()

        return osmc_response.json()

    def element_short_string(self, element_descriptor):
        raw_container_content, raw_element_content = self.retrieve(element_descriptor)
        return _twc_element_short_string(raw_container_content, raw_element_content)
        
    def show_element_info_widget(self, element_descriptor, **kwargs):
        return TWCElementInfoWidget(osmc=self, initial_element=element_descriptor, **kwargs)

def _twc_element_short_string(raw_container_content, raw_element_content):
    # nsUri = raw_element_content["kerml:nsURI"]
    qualified_type = raw_element_content.get("@type", "(untyped)")#.split(':')
    #meta_package = qualified_type[0]
    #type = qualified_type[1]
    name = raw_element_content.get("kerml:name", None)
    _id = raw_element_content.get("kerml:esiID", '<unknown ID>')
    if name:
        return "'{}' ({}) #{}".format(name, qualified_type, _id)
    else:
        return "<unnamed> ({}) #{}".format(qualified_type, _id)



class TWCElementInfoWidget(abstract_widgets.AbstractElementInfoWidget):
    def __init__(
        self,
        osmc,
        initial_element,
        **kwargs
    ):
        self.osmc = osmc
        self.initial_element = initial_element
        self.element_id_to_descriptor = {}
        super().__init__(self, **kwargs)
        self.show_element_pane(initial_element)
    
    def _unpack_element(self, element) -> Tuple[str, Dict[str, Any]]:
        raw_container_content, raw_element_content = self.osmc.retrieve(element)
        header = _twc_element_short_string(raw_container_content, raw_element_content)
        field_values = raw_element_content.get('kerml:esiData', {})
        return header, field_values
    
    def _do_retrieve_batch(self, element, field_values):
        self.osmc.retrieve_batch([element_descriptor
            for field,value in field_values.items()
            if (value)
            for element_descriptor in self._value_element_descriptor_list(value)
        ])
    def _value_element_descriptor_list(self, value):
        import collections
        if isinstance(value, str):
            return []
        elif isinstance(value, collections.abc.Sequence): # test string-ness before this
            flat_list_descriptors = [ descriptor
                for item in value 
                for descriptor in self._value_element_descriptor_list(item)
            ]
            return flat_list_descriptors
        elif isinstance(value, dict) and '@id' in value:
            referred_element = self._element_descriptor_for_id(value['@id'])
            return [referred_element]
        else:
            return []

    def _to_hashable_identifier(self, element) -> Hashable:
        return element.element_id

    def _to_field_display_name(self, element, field) -> str:
        return field

    def _value_widget_list(self, element, field, value) -> List[Widget]:
        import collections
        if isinstance(value, str):
            return [self._create_value_widget_attribute(value)]#'"{}"'.format(value))
        elif isinstance(value, collections.abc.Sequence): # test string-ness before this
            flat_list_widgets = [ widget
                for item in value 
                for widget in self._value_widget_list(element, field, item)
            ]
            return flat_list_widgets
        elif isinstance(value, dict) and '@id' in value:
            referred_element = self._element_descriptor_for_id(value['@id'])
            short_text = self.osmc.element_short_string(referred_element)
            long_text = 'Click to open pane for element {} found at {}'.format(short_text, referred_element.to_descriptor_str())
            return [self._create_value_widget_reference(referred_element, short_text, long_text)]
        else:
            return [self._create_value_widget_attribute(value)]

    def _element_descriptor_for_id(self, element_id):
        if element_id not in self.element_id_to_descriptor:
            self.element_id_to_descriptor[element_id] = self.initial_element.resolve_reference(element_id)
        return self.element_id_to_descriptor[element_id]




# Copyright 2019 IncQuery Labs Ltd.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#  http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Created on 2019. máj. 24.

@author: Gabor Bergmann
'''

from typing import Optional
import ipywidgets as widgets
from IPython.display import display

import iqs_client

import iqs_jupyter.config_defaults as defaults
import iqs_jupyter.tool_extension_point as ext_point



class TWCRevisionInputWidget:
    def __init__(
        self,
        initial_workspace = defaults.default_twc_workspace,
        initial_resource  = defaults.default_twc_resource,
        initial_branch    = defaults.default_twc_branch,
        initial_revision  = defaults.default_twc_revision,
        auto_display=True
    ):
        self.workspace_widget = widgets.Text(value=initial_workspace,   description='ID:')
        self.resource_widget  = widgets.Text(value=initial_resource,    description='ID:')
        self.branch_widget    = widgets.Text(value=initial_branch,      description='ID:')
        self.revision_widget  = widgets.IntText(value=initial_revision, description='Number:')
        self.box=widgets.HBox([
            widgets.VBox([widgets.Label(value="Workspace"), widgets.Label(value="Resource"), widgets.Label(value="Branch"), widgets.Label(value="Revision")]),
            widgets.VBox([self.workspace_widget,            self.resource_widget,            self.branch_widget,            self.revision_widget])
        ])
        if auto_display: self.display()
    
    def display(self):
        display(self.box)
        
    def _repr_html_(self):
        self.display()    
        
    def value(self):
        return iqs_client.RevisionDescriptor(
            revision_number = self.revision_widget.value,
            branch_id       = self.branch_widget.value,
            resource_id     = self.resource_widget.value,
            workspace_id    = self.workspace_widget.value
        )


class TWCRevisionSelectorWidget:
    
    def __init__(
        self,
        iqs,
        initial_workspace = defaults.default_twc_workspace,
        initial_resource  = defaults.default_twc_resource,
        initial_branch    = defaults.default_twc_branch,
        initial_revision  = defaults.default_twc_revision,
        auto_display=True
    ):
        self.iqs = iqs
        self.workspace_widget = widgets.Dropdown(description='')
        self.resource_widget  = widgets.Dropdown(description='')
        self.branch_widget    = widgets.Dropdown(description='')
        self.revision_widget  = widgets.Dropdown(description='Number:')
        self.box=widgets.HBox([
            widgets.VBox([widgets.Label(value="Workspace"), widgets.Label(value="Resource"), widgets.Label(value="Branch"), widgets.Label(value="Revision")]),
            widgets.VBox([self.workspace_widget,            self.resource_widget,            self.branch_widget,            self.revision_widget])
        ])
        self.fetch_contents()
        
        def workspace_handler(change):
            new_idx = change['new']
            old_idx = change['old']
            owner = change['owner']
            if (new_idx == old_idx):
                pass
            elif (not new_idx):
                self._reset_resource()
            elif (owner.disabled):
                pass
            else:
                label, _id = owner.options[new_idx]  # @UnusedVariable
                workspace = self.workspace_map[_id]
                self._setup_resource(workspace.resources)

        def resource_handler(change):
            new_idx = change['new']
            old_idx = change['old']
            owner = change['owner']
            if (new_idx == old_idx):
                pass
            elif (not new_idx):
                self._reset_branch()
            elif (owner.disabled):
                pass
            else:
                label, _id = self.resource_widget.options[new_idx]  # @UnusedVariable
                resource = self.resource_map[_id]
                self._setup_branch(resource.branches)
                
        def branch_handler(change):
            new_idx = change['new']
            old_idx = change['old']
            owner = change['owner']
            if (new_idx == old_idx):
                pass
            elif (not new_idx):
                self._reset_revision()
            elif (owner.disabled):
                pass
            else:
                label, _id = self.branch_widget.options[new_idx]  # @UnusedVariable
                branch = self.branch_map[_id]
                self._setup_revision(branch.revisions)
                
        self.workspace_widget.observe(names='index', handler = workspace_handler)
        self.resource_widget.observe(names='index', handler = resource_handler)
        self.branch_widget.observe(names='index', handler = branch_handler)
        
        if initial_workspace:
            self.workspace_widget.value = initial_workspace
            if initial_resource:
                self.resource_widget.value = initial_resource
                if initial_branch:
                    self.branch_widget.value = initial_branch
                    if initial_revision:
                        self.revision_widget.value = initial_revision
        
        if auto_display: self.display()
    
    
    
    def _reset_resource(self):
        self.resource_map = None
        self.resource_widget.disabled = True
        self.resource_widget.options=[('', None)]
        self.resource_widget.index=0
        #self._reset_branch()
        
    def _reset_branch(self):
        self.branch_map = None
        self.branch_widget.disabled = True
        self.branch_widget.options=[('', None)]
        self.branch_widget.index=0
        #self._reset_revision()
        
    def _reset_revision(self):
        self.revision_map = None
        self.revision_widget.disabled = True
        self.revision_widget.options=[('', None)]
        self.revision_widget.index=0
        
    def _setup_workspace(self, workspace_list):
        import collections
        self.workspace_map = collections.OrderedDict(
            [(workspace.workspace_id, workspace) for workspace in workspace_list])
        self.workspace_widget.options=[('---- Select workspace', None)] + [
                (
                    "{} (ID: {})".format(workspace.title, _id),
                    _id
                ) 
                for _id,workspace in self.workspace_map.items()
            ]
        self.workspace_widget.index=0
        self.workspace_widget.disabled = False
        #self._reset_resource()
        
    def _setup_resource(self, resource_list):
        import collections
        self.resource_map = collections.OrderedDict([(resource.resource_id, resource) for resource in resource_list])
        self.resource_widget.options=[('---- Select resource', None)] + [
                (
                    "{} (ID: {})".format(resource.title, _id),
                    _id
                ) 
                for _id,resource in self.resource_map.items()
            ]
        self.resource_widget.index=0
        self.resource_widget.disabled = False
        #self._reset_branch()
        
    def _setup_branch(self, branch_list):
        import collections
        self.branch_map = collections.OrderedDict([(branch.branch_id, branch) for branch in branch_list])
        self.branch_widget.options=[('---- Select branch', None)] + [
                (
                    "{} (ID: {})".format(branch.title, _id),
                    _id
                ) 
                for _id,branch in self.branch_map.items()
            ]
        self.branch_widget.index=0
        self.branch_widget.disabled = False
        #self._reset_revision()
        
    def _setup_revision(self, revision_list):
        import collections
        self.revision_map = collections.OrderedDict([ (revision['revisionNumber'], revision) for revision in revision_list ])
        self.revision_widget.options=[('---- Select revision', None)] + [
                (str(number), number)
                for number,revision in self.revision_map.items()
            ]
        self.revision_widget.index=0
        self.revision_widget.disabled = False
        
    def fetch_contents(self):
        repo_info = self.iqs.repository.get_repository_info()
        workspace_list = repo_info.repository_structure.workspaces
        self._setup_workspace(workspace_list)

    def display(self):
        display(self.box)
        
    def _repr_html_(self):
        self.display()    
        
    def value(self):
        if (self.revision_widget.index  != 0 and 
            self.branch_widget.index    != 0 and 
            self.resource_widget.index  != 0 and 
            self.workspace_widget.index != 0 
        ):
            return iqs_client.RevisionDescriptor(
                revision_number = self.revision_widget.value,
                branch_id       = self.branch_widget.value,
                resource_id     = self.resource_widget.value,
                workspace_id    = self.workspace_widget.value
            )
        else: 
            return None



# recognizing element descriptors in match results

def _recognize_element_descriptor(dict_of_element : dict) -> Optional[iqs_client.ElementDescriptor]:
    if "workspaceId" in dict_of_element:
        return iqs_client.ElementDescriptor(
            **dict(
                (py_name, dict_of_element[json_name]) 
                for py_name, json_name in iqs_client.ElementDescriptor.attribute_map.items()
            )
        )
    else:
        return None
        
ext_point.element_dict_recognizers.append(_recognize_element_descriptor)


# monkey patch section

def _monkey_patch_element_descriptor_to_str(self):
    return 'element #{} at {}'.format(self.element_id, self.to_descriptor_str())
    
def _monkey_patch_element_descriptor_to_descriptor_str(self):
    import pprint
    return pprint.pformat(self.to_dict())

def _monkey_patch_element_descriptor_repr_html(self):
    import html
    if hasattr(self, 'url'):
        return '<a href="{}" title="{}">element #{}</a>'.format(html.escape(self.url), html.escape(self), html.escape(self.element_id))
    else:
        return '<span title="{}">element #{}</span>'.format(html.escape(self.to_str()), html.escape(self.element_id))

def _monkey_patch_element_descriptor_resolve_reference(self, target_element_id):
    return iqs_client.ElementDescriptor(
        revision_number = self.revision_number,
        branch_id       = self.branch_id,
        resource_id     = self.resource_id,
        workspace_id    = self.workspace_id,
        element_id      = target_element_id
    )
    
def _monkey_patch_element_descriptor_get_containing_revision(self):
    return iqs_client.RevisionDescriptor(
        revision_number = self.revision_number,
        branch_id       = self.branch_id,
        resource_id     = self.resource_id,
        workspace_id    = self.workspace_id
    )
    
def _monkey_patch_revision_descriptor_get_element_descriptor_by_id(self, element_id):
    return iqs_client.ElementDescriptor(
        revision_number = self.revision_number,
        branch_id       = self.branch_id,
        resource_id     = self.resource_id,
        workspace_id    = self.workspace_id,
        element_id      = element_id
    )
    
def _monkey_patch_revision_descriptor_to_compartment_uri(self):
    return "twc-index:/workspaces/{}/resources/{}/branches/{}/revisions/{}".format(
        self.workspace_id, self.resource_id, self.branch_id, self.revision_number
    )


def _monkey_patch_revision_descriptor_to_model_compartment(self):
    return iqs_client.ModelCompartment(compartment_uri=self.to_compartment_uri())

def _monkey_patch_jupytertools_twc_revision_selector_widget(self, **kwargs):
    return TWCRevisionSelectorWidget(iqs=self._iqs, **kwargs)

def _do_monkey_patching():
    iqs_client.ElementDescriptor.to_str = _monkey_patch_element_descriptor_to_str
    iqs_client.ElementDescriptor.to_descriptor_str = _monkey_patch_element_descriptor_to_descriptor_str
    iqs_client.ElementDescriptor._repr_html_ = _monkey_patch_element_descriptor_repr_html
    iqs_client.ElementDescriptor.resolve_reference = _monkey_patch_element_descriptor_resolve_reference
    iqs_client.ElementDescriptor.get_containing_revision = _monkey_patch_element_descriptor_get_containing_revision
    iqs_client.RevisionDescriptor.get_element_descriptor_by_id = _monkey_patch_revision_descriptor_get_element_descriptor_by_id
    iqs_client.RevisionDescriptor.to_compartment_uri = _monkey_patch_revision_descriptor_to_compartment_uri
    iqs_client.RevisionDescriptor.to_model_compartment = _monkey_patch_revision_descriptor_to_model_compartment
    
    ext_point.IQSJupyterTools.twc_revision_selector_widget = _monkey_patch_jupytertools_twc_revision_selector_widget

_do_monkey_patching()



# Copyright 2019 IncQuery Labs Ltd.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#  http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
    IncQuery Server Client Extensions for Jupyter and Python

    Requires [IncQuery Server](https://incquery.io) to operate.


Created on 2019-05-17

@author: Gábor Bergmann
'''

# coding: utf-8
from iqs_jupyter.about import __version__

# end-user modules
from iqs_jupyter.core_extensions import *

if 'MmsRepositoryApi' in dir(iqs_client):
    from iqs_jupyter.mms_extensions import *
    from iqs_jupyter.mms_direct_extensions import *
    
if 'RepositoryApi' in dir(iqs_client):
    from iqs_jupyter.twc_extensions import *
    from iqs_jupyter.twc_osmc_extensions import *

if 'AnalysisApi' in dir(iqs_client):
    from iqs_jupyter.analysis_extensions import *

# namespace shortcut for API request/response classes
from iqs_client import models as schema





# Copyright 2019 IncQuery Labs Ltd.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#  http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Created on 2019. máj. 21.

@author: Gabor Bergmann
"""

_prefix = "IQS_JUPYTER_"

default_IQS_address  : str = None
default_IQS_username : str = None
default_IQS_password : str = None

default_twc_workspace : str = None
default_twc_resource  : str = None
default_twc_branch    : str = None
default_twc_revision  : int = None

default_mms_org     : str = None
default_mms_project : str = None
default_mms_ref     : str = None
default_mms_commit  : str = None

default_twc_osmc_address  : str = None
default_twc_osmc_username : str = None
default_twc_osmc_password : str = None

default_mms_address  : str = None
default_mms_username : str = None
default_mms_password : str = None


# try to override from env variables

    
# for var_name, default in [ # copy to avoid concurrent iteration
#     (var_name, default) 
#     for var_name, default in globals().items()
#     if var_name.startswith("default_")
# ]:
for var_name, default in [ # copy to avoid concurrent modification
    (var_name, default) 
    for var_name, default in globals().items()
    if var_name.startswith("default_")
]:
    from os import environ
    from typing import get_type_hints
    from sys import modules
    value = environ.get(_prefix + var_name)
    hint = get_type_hints(modules[__name__]).get(var_name)
    if value is None:
        globals()[var_name] = default
    else: 
        if hint == int:
            globals()[var_name] = int(value)
        else:
            globals()[var_name] = value




# Copyright 2019 IncQuery Labs Ltd.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#  http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
'''
Created on 2019. nov. 26.

@author: Gabor Bergmann
'''

import html

import iqs_client
from iqs_jupyter.helpers import cell_to_html as _cell_to_html
from iqs_jupyter.helpers import dict_to_element as _dict_to_element
from iqs_jupyter.core_extensions import validation_color_scale


def _monkey_patch_analysis_results_repr_html(self : iqs_client.AnalysisResults):
    escaped_id = html.escape(self.configuration.configuration_id)
    header_report = '''
        <h3>Results for model analysis <b>"{}"</b></h3>
        <span title="{}">Analysis configuration id: {} <i>(see hover for details)</i></span>
    '''.format(
        html.escape(self.configuration.configuration_name),
        html.escape(self.to_str()), 
        escaped_id
    )
    
    kpi_rows = "\n".join([
        "<tr><th>{}</th><td>{}</td></tr>".format(
            html.escape(kpi.name), 
            html.escape(str(kpi.value))
        ) 
        for kpi in self.kpi_values
    ])
    kpi_report = '''
        <h4> KPIs</h4>
        <table border="1">
            <thead>
                <tr style="text-align: right;">
                    <th>KPI</th>
                    <th>Value</th>
                </tr>
            </thead>
            <tbody>
                {}
            </tbody>
        </table>
    '''.format(
        kpi_rows
    )
    
    toc_rows = "\n".join([
        '''
            <li>{} by rule "<b>{}</b>"</li>
        '''.format(
            _marker_count_report(self.analysis_results[result_index]),
            #escaped_id, str(result_index),
            html.escape(self.analysis_results[result_index].configuration_rule.name)
        ) 
        for result_index in range(len(self.analysis_results))
    ])
    toc_report = '''
        <h4>Individual analysis rules</h4>
        <ul>
            {}
        </ul>
    '''.format(
        toc_rows
    )

    
    individual_result_tables = "\n".join([
        '''
            <div>
            {}
            </div>
        '''.format(
            self.analysis_results[result_index]._repr_html_("h4")
        ) 
        for result_index in range(len(self.analysis_results))
    ])
    
    style = ""
    return '''
        {}
        <div>
        {}
        {}
        {}
        {}
        </div>
    '''.format(header_report, style, kpi_report, toc_report, individual_result_tables)


def _marker_count_report(analysis_result : iqs_client.AnalysisResult) -> str:
    marker_color = validation_color_scale.get(analysis_result.configuration_rule.severity.lower(), None)
    marker_style_string = 'style="background-color:{}; color: bisque;"'.format(marker_color) if marker_color else ""
    return '{} <span {}>{}</span> marker{}'.format(
        str(len(analysis_result.matches)),
        marker_style_string,
        html.escape(analysis_result.configuration_rule.severity),
        "s" if 1 != len(analysis_result.matches) else ""
    )

def _monkey_patch_analysis_result_repr_html(self : iqs_client.AnalysisResult, heading_lvl: str = 'h4', anchor : str = None):
    anchor_tag = '<a name="{}"/>'.format(anchor) if anchor else ""
    anchor_href = '<a href="#{}">#</a> '.format(anchor) if anchor else ""
    subheader_report = '{}<span title="{}">{} via "{}" <i>(see hover for details)</i></span>'.format(
        anchor_href,
        html.escape(self.to_str()), 
        _marker_count_report(self),
        html.escape(self.configuration_rule.query_fqn)
    )
    header_report = '''
        {}
        <{}>Results of rule <b>"{}"</b></{}>
        {}
    '''.format(
        anchor_tag,
        heading_lvl,
        html.escape(self.configuration_rule.name),
        heading_lvl,
        subheader_report
    )
    if not self.matches:
        return header_report
    
    # we can assume at least one match
    pattern_params = [
        argument_value.parameter
        for argument_value in self.matches[0].matching_elements 
    ]
    
    param_headers = " ".join(["<th>{}</th> ".format(html.escape(param)) for param in pattern_params])
    match_rows = "\n".join([
        "<tr><th>{}</th><td>{}</td>{}</tr>\n".format(row_index, self.matches[row_index].message, " ".join([
            " <td>{}</td>".format(_cell_to_html(_dict_to_element(argument_value.value))) 
            for argument_value in self.matches[row_index].matching_elements
        ])) for row_index in range(len(self.matches))
    ])
    
    style = ""
    return '''
        {}
        <div>
        {}
        <table border="1">
            <thead>
                <tr style="text-align: right;">
                    <th></th>
                    <th>(Message)</th>
                    {}
                </tr>
            </thead>
            <tbody>
                {}
            </tbody>
        </table>
        </div>
    '''.format(header_report, style, param_headers, match_rows)



def _do_monkey_patching():
    iqs_client.AnalysisResults._repr_html_ = _monkey_patch_analysis_results_repr_html
    iqs_client.AnalysisResult._repr_html_ = _monkey_patch_analysis_result_repr_html

_do_monkey_patching()



# Copyright 2019 IncQuery Labs Ltd.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#  http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Created on 2019. máj. 24.

@author: Gabor Bergmann
'''

import ipywidgets as widgets
from IPython.display import display

import iqs_client

import iqs_jupyter.config_defaults as defaults
import iqs_jupyter.tool_extension_point as ext_point


class MMCCommitSelectorWidget:

    def __init__(
            self,
            iqs,
            initial_org     = defaults.default_mms_org,
            initial_project = defaults.default_mms_project,
            initial_ref     = defaults.default_mms_ref,
            initial_commit  = defaults.default_mms_commit,
            auto_display=True
    ):
        self.iqs = iqs
        self.org_widget = widgets.Dropdown(description='')
        self.project_widget = widgets.Dropdown(description='')
        self.ref_widget = widgets.Dropdown(description='')
        self.commit_widget = widgets.Dropdown(description='')
        self.box = widgets.HBox([
            widgets.VBox([widgets.Label(value="org"), widgets.Label(value="project"), widgets.Label(value="ref"),
                          widgets.Label(value="commit")]),
            widgets.VBox([self.org_widget, self.project_widget, self.ref_widget, self.commit_widget])
        ])
        self.fetch_contents()

        def org_handler(change):
            new_idx = change['new']
            old_idx = change['old']
            owner = change['owner']
            if (new_idx == old_idx):
                pass
            elif (not new_idx):
                self._reset_project()
            elif (owner.disabled):
                pass
            else:
                label, _id = owner.options[new_idx]  # @UnusedVariable
                org = self.org_map[_id]
                self._setup_project(org.projects)

        def project_handler(change):
            new_idx = change['new']
            old_idx = change['old']
            owner = change['owner']
            if (new_idx == old_idx):
                pass
            elif (not new_idx):
                self._reset_ref()
            elif (owner.disabled):
                pass
            else:
                label, _id = self.project_widget.options[new_idx]  # @UnusedVariable
                project = self.project_map[_id]
                self._setup_ref(project.refs)

        def ref_handler(change):
            new_idx = change['new']
            old_idx = change['old']
            owner = change['owner']
            if (new_idx == old_idx):
                pass
            elif (not new_idx):
                self._reset_commit()
            elif (owner.disabled):
                pass
            else:
                label, _id = self.ref_widget.options[new_idx]  # @UnusedVariable
                ref = self.ref_map[_id]
                self._setup_commit(ref.commits)

        self.org_widget.observe(names='index', handler=org_handler)
        self.project_widget.observe(names='index', handler=project_handler)
        self.ref_widget.observe(names='index', handler=ref_handler)

        if initial_org:
            self.org_widget.value = initial_org
            if initial_project:
                self.project_widget.value = initial_project
                if initial_ref:
                    self.ref_widget.value = initial_ref
                    if initial_commit:
                        self.commit_widget.value = initial_commit

        if auto_display: self.display()

    def _reset_project(self):
        self.project_map = None
        self.project_widget.disabled = True
        self.project_widget.options = [('', None)]
        self.project_widget.index = 0

    # self._reset_ref()

    def _reset_ref(self):
        self.ref_map = None
        self.ref_widget.disabled = True
        self.ref_widget.options = [('', None)]
        self.ref_widget.index = 0

    # self._reset_commit()

    def _reset_commit(self):
        self.commit_map = None
        self.commit_widget.disabled = True
        self.commit_widget.options = [('', None)]
        self.commit_widget.index = 0

    def _setup_org(self, org_list):
        import collections
        org_list_sorted = sorted(org_list, key=lambda org: org.name)
        self.org_map = collections.OrderedDict(
            [(org.org_id, org) for org in org_list_sorted])
        self.org_widget.options = [('---- Select org', None)] + [
            (
                "{} (ID: {})".format(org.name, _id),
                _id
            )
            for _id, org in self.org_map.items()
        ]
        self.org_widget.index = 0
        self.org_widget.disabled = False

    # self._reset_project()

    def _setup_project(self, project_list):
        import collections
        project_list_sorted = sorted(project_list, key=lambda project: project.name)
        self.project_map = collections.OrderedDict([(project.project_id, project) for project in project_list_sorted])
        self.project_widget.options = [('---- Select project', None)] + [
            (
                "{} (ID: {})".format(project.name, _id),
                _id
            )
            for _id, project in self.project_map.items()
        ]
        self.project_widget.index = 0
        self.project_widget.disabled = False

    # self._reset_ref()

    def _setup_ref(self, ref_list):
        import collections
        ref_list_sorted = sorted(ref_list, key=lambda ref: ref.name)
        self.ref_map = collections.OrderedDict([(ref.ref_id, ref) for ref in ref_list_sorted])
        self.ref_widget.options = [('---- Select ref', None)] + [
            (
                "{} (ID: {})".format(ref.name, _id),
                _id
            )
            for _id, ref in self.ref_map.items()
        ]
        self.ref_widget.index = 0
        self.ref_widget.disabled = False

    # self._reset_commit()

    def _setup_commit(self, commit_list):
        import collections
        commit_list_sorted = sorted(commit_list, key=lambda commit: commit['name'], reverse=True)
        self.commit_map = collections.OrderedDict([(commit['commitId'], commit) for commit in commit_list_sorted])
        self.commit_widget.options = [('---- Select commit', None)] + [
            (
                "{} (ID: {})".format(commit['name'], _id),
                _id
            )
            for _id, commit in self.commit_map.items()
        ]
        self.commit_widget.index = 0
        self.commit_widget.disabled = False

    def fetch_contents(self):
        repo_info = self.iqs.mms_repository.get_mms_repository_info()
        org_list = repo_info.repository_structure.orgs
        self._setup_org(org_list)

    def display(self):
        display(self.box)

    def _repr_html_(self):
        self.display()

    def value(self):
        if (self.commit_widget.index != 0 and
                self.ref_widget.index != 0 and
                self.project_widget.index != 0 and
                self.org_widget.index != 0
        ):
            return iqs_client.MMSCommitDescriptor.from_fields(
                name=self.commit_map[self.commit_widget.value]['name'],
                commit_id=self.commit_widget.value,
                ref_id=self.ref_widget.value,
                project_id=self.project_widget.value,
                org_id=self.org_widget.value
            )
        else:
            return None

    def value_as_model_compartment(self):
        value = self.value()
        if value:
            return value.to_model_compartment()
        else:
            return None

# monkey patch section


def _monkey_patch_static_mms_commit_descriptor_from_compartment_uri_or_none(compartment_uri):
    segments = compartment_uri.split('/')
    if (
        9 != len(segments) or
        segments[0] != "mms-index:" or
        segments[1] != "orgs"       or
        segments[3] != "projects"   or
        segments[5] != "refs"       or
        segments[7] != "commits"    
    ):
        return None
    
    return iqs_client.MMSCommitDescriptor(
        org_id          = segments[2], 
        project_id      = segments[4], 
        ref_id          = segments[6], 
        commit_id       = segments[8],
        compartment_uri = compartment_uri
    )
    
def _monkey_patch_mms_commit_to_compartment_uri(self):
    return "mms-index:/orgs/{}/projects/{}/refs/{}/commits/{}".format(
        self.org_id,
        self.project_id,
        self.ref_id,
        self.commit_id
    )
def _mms_compartment_uri(org_id, project_id, ref_id, commit_id):
    return "mms-index:/orgs/{}/projects/{}/refs/{}/commits/{}".format(
        org_id,
        project_id,
        ref_id,
        commit_id
    )
def _monkey_patch_static_mms_commit_descriptor_from_fields(org_id, project_id, ref_id, commit_id, name = None):
    return iqs_client.MMSCommitDescriptor(
        name=name,
        commit_id=commit_id,
        ref_id=ref_id,
        project_id=project_id,
        org_id=org_id,
        compartment_uri=_mms_compartment_uri(org_id, project_id, ref_id, commit_id)
    )

def _monkey_patch_mms_commit_to_model_compartment(self):
    return iqs_client.ModelCompartment(compartment_uri=self.to_compartment_uri())


def _monkey_patch_jupytertools_mms_commit_selector_widget(self, **kwargs):
    return MMCCommitSelectorWidget(iqs=self._iqs, **kwargs)

def _do_monkey_patching():
    iqs_client.MMSCommitDescriptor.from_compartment_uri_or_none = staticmethod(_monkey_patch_static_mms_commit_descriptor_from_compartment_uri_or_none)
    iqs_client.MMSCommitDescriptor.from_fields = staticmethod(_monkey_patch_static_mms_commit_descriptor_from_fields)
    iqs_client.MMSCommitDescriptor.to_compartment_uri = _monkey_patch_mms_commit_to_compartment_uri
    iqs_client.MMSCommitDescriptor.to_model_compartment = _monkey_patch_mms_commit_to_model_compartment
    
    ext_point.IQSJupyterTools.mms_commit_selector_widget = _monkey_patch_jupytertools_mms_commit_selector_widget

_do_monkey_patching()



# Copyright 2019 IncQuery Labs Ltd.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#  http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Dict, Tuple, Hashable, List
from ipywidgets.widgets.widget import Widget
'''
Created on 2019. júl. 4.

@author: Gabor Bergmann
'''

import mms_python_client as mms_python_client

import iqs_client

import iqs_jupyter.config_defaults as defaults
from iqs_jupyter import abstract_widgets



_mms_client_api_classes = {
    'artifact'   : 'ArtifactApi',     
    'element'    : 'ElementApi',      
    'org'        : 'OrgApi',          
    'other'      : 'OtherApi',        
    'project'    : 'ProjectApi',      
    'ref'        : 'RefApi',          
    'ticket'     : 'TicketApi'        
}


class MMSClient:
    def __init__(
        self, 
        address  = defaults.default_mms_address,
        user     = defaults.default_mms_username,
        password = defaults.default_mms_password,
        cache_element_details_locally = True
    ):
        if (address.endswith('/')):
            self.address = address
        else:
            self.address = address + '/'

        configuration = mms_python_client.Configuration( )
        configuration.host=address 
        configuration.username=user
        configuration.password=password
        
        core_api_client_mms = mms_python_client.ApiClient(configuration)
        for api_field_name, api_class_name in _mms_client_api_classes.items():
            if api_class_name in dir(mms_python_client): # there might be api classes presently missing / disabled
                api_class = getattr(mms_python_client, api_class_name)
                api_object = api_class(core_api_client_mms)
                setattr(self, api_field_name, api_object)
        # get_elements_in_batch
        
        self.helpers = MMSHelpers(self, cache_element_details_locally)
        
        
class MMSHelpers:
    RAW_ELEMENT_CONTENT='_raw_element_content'
    
    # not to be treated as in-model cross-references
    META_REF_FIELDS={'_commitId','_elasticId','_refId','_inRefIds','_projectId','mountedElementId','mountedElementProjectId','mountedRefId'} 
    
    def __init__(
        self, 
        mms,
        cache_locally = True
    ):
        self.mms = mms
        self.cache_locally = cache_locally
        
    def retrieve(self, element):
        if hasattr(element, MMSHelpers.RAW_ELEMENT_CONTENT):
            return getattr(element, MMSHelpers.RAW_ELEMENT_CONTENT)
        else:
            return self.force_retrieve(element)

    def force_retrieve(self, element):
        raw_element_content = self._retrieve_single(element)
        if self.cache_locally:
            setattr(element, MMSHelpers.RAW_ELEMENT_CONTENT, raw_element_content)
        return raw_element_content

    def _retrieve_single(self, element):
        commit = iqs_client.MMSCommitDescriptor.from_compartment_uri_or_none(element.compartment_uri)
        if not commit:
            return None
            
        try:
            elements_response = self.mms.element.get_element(
                project_id=commit.project_id, 
                ref_id=commit.ref_id, 
                commit_id=commit.commit_id,
                element_id=element.relative_element_id,    
            )
            if (elements_response.elements):
                return elements_response.elements[0].to_dict()
            else:
                return None
        except:
            return None
    
        
    def retrieve_batch(self, element_in_compartment_list):
        if self.cache_locally:
            action_needed = [
                element 
                for element in element_in_compartment_list 
                if not (hasattr(element, MMSHelpers.RAW_ELEMENT_CONTENT))
            ]
            if action_needed:
                self.force_retrieve_batch(action_needed)

    def force_retrieve_batch(self, element_in_compartment_list):
        from itertools import groupby
        grouping = groupby(element_in_compartment_list, lambda element: element.compartment_uri)
        for compartment_uri, group_with_same_compartment in grouping:
            raw_content_dict = self._retrieve_batch(compartment_uri, group_with_same_compartment)
            if self.cache_locally:
                for element in element_in_compartment_list:
                    if element.relative_element_id in raw_content_dict:
                        raw_element_content = raw_content_dict[element.relative_element_id]
                        setattr(element, MMSHelpers.RAW_ELEMENT_CONTENT, raw_element_content)

    def _retrieve_batch(self, compartment_uri, element_in_compartment_list):
        commit = iqs_client.MMSCommitDescriptor.from_compartment_uri_or_none(compartment_uri)
        if not commit:
            return {}
            
        rejectable_elements_response = self.mms.element.get_elements_in_batch(
            project_id=commit.project_id, 
            ref_id=commit.ref_id, 
            commit_id=commit.commit_id,
            body={'elements':[
                {'id': element.relative_element_id}
                for element in element_in_compartment_list
            ]},    
        )
        return {
            raw_element.id: raw_element.to_dict()
            for raw_element in rejectable_elements_response.elements
        }
        
    def element_short_string(self, element):
        raw_element_content = self.retrieve(element)
        return _mms_element_short_string(element, raw_element_content)
        
    def show_element_info_widget(self, 
                                 element, #element_id or element_in_compartment
                                 org_id=None,
                                 project_id=None,
                                 ref_id=None,
                                 commit_id=None, 
                                 **kwargs):
        if (org_id and project_id and ref_id and commit_id):
            commit = iqs_client.MMSCommitDescriptor.from_fields(
                org_id=org_id, project_id=project_id, ref_id=ref_id, commit_id=commit_id
            )
            element = iqs_client.ElementInCompartmentDescriptor(
                compartment_uri = commit.to_compartment_uri(),
                relative_element_id = element  # assume `element` is relative element id
            )            
        
        return MMSElementInfoWidget(mms_helper=self, initial_element=element, **kwargs)
        

def _mms_element_short_string(element, raw_element_content):
    if not raw_element_content:
        return "<missing element> #{}".format(element.relative_element_id)

    # nsUri = raw_element_content["kerml:nsURI"]
    qualified_type = raw_element_content.get("type", "<untyped>")#.split(':')
    #meta_package = qualified_type[0]
    #type = qualified_type[1]
    name = raw_element_content.get("name", None)
    _id = element.relative_element_id #raw_element_content.get("id", '<unknown ID>')
    if name:
        return "'{}' ({}) #{}".format(name, qualified_type, _id)
    else:
        return "<unnamed> ({}) #{}".format(qualified_type, _id)

class MMSElementInfoWidget(abstract_widgets.AbstractElementInfoWidget):
    def __init__(
        self,
        mms_helper,
        initial_element,
        **kwargs
    ):
        self.mms_helper = mms_helper
        self.initial_element = initial_element
        self.element_id_to_element = {}
        super().__init__(self, **kwargs)
        self.show_element_pane(initial_element)
    
    def _unpack_element(self, element) -> Tuple[str, Dict[str, Any]]:
        raw_element_content = self.mms_helper.retrieve(element)
        header = _mms_element_short_string(element, raw_element_content)
        field_values = raw_element_content if raw_element_content else {}
        return header, field_values
    
    def _do_retrieve_batch(self, element, field_values):
        self.mms_helper.retrieve_batch([element_ref
            for field,value in field_values.items()
            if (value)
            for element_ref in self._value_element_ref_list(element, field, value)
        ])
    def _value_element_ref_list(self, element, field, value):
        if field.endswith("Id") and not field in MMSHelpers.META_REF_FIELDS:
            return [self._element_for_id(value)]
        elif field.endswith("Ids") and not field in MMSHelpers.META_REF_FIELDS:
            return [
                self._element_for_id(item)
                for item in value
            ]
        #elif isinstance(value, dict) and 'id' in value:
        #    referred_element = self._element_for_id(value['id'])
        #    return [referred_element]
        else:
            return []

    def _to_hashable_identifier(self, element) -> Hashable:
        return element.relative_element_id

    def _to_field_display_name(self, element, field) -> str:
        if field.endswith("Id"):
            return field[:-2]
        elif field.endswith("Ids"):
            return field[:-3]
        else:
            return field
        
    def _value_widget_list(self, element, field, value) -> List[Widget]:
        import collections
        if field.endswith("Id") and not field in MMSHelpers.META_REF_FIELDS:
            return [self._ref_widget(element, value)]
        elif field.endswith("Ids") and not field in MMSHelpers.META_REF_FIELDS:
            return [
                self._ref_widget(element, item)
                for item in value
            ]
        elif isinstance(value, str):
            return [self._create_value_widget_attribute(value)]#'"{}"'.format(value))
        elif isinstance(value, collections.abc.Sequence): # test string-ness before this
            return [
                self._ref_or_attrib_widget(element, item) 
                for item in value 
            ]
        else:
            return [self._ref_or_attrib_widget(element, value)]
        
    def _ref_widget(self, src_element, value):
        referred_element = self._element_for_id(value)
        short_text = self.mms_helper.element_short_string(referred_element)
        long_text = 'Click to open pane for element {}'.format(short_text)
        return self._create_value_widget_reference(referred_element, short_text, long_text)

    def _ref_or_attrib_widget(self, element, value):
        if isinstance(value, dict) and 'id' in value:
            _id = value['id']
            referred_element = self._element_for_id(_id)
            if not hasattr(referred_element, MMSHelpers.RAW_ELEMENT_CONTENT):
                setattr(referred_element, MMSHelpers.RAW_ELEMENT_CONTENT, value)
            return self._ref_widget(element, _id)
        else:
            return self._create_value_widget_attribute(value)
    
    

    def _element_for_id(self, element_id):
        if element_id not in self.element_id_to_element:
            self.element_id_to_element[element_id] = self.initial_element.resolve_reference(element_id)
        return self.element_id_to_element[element_id]




# Copyright 2019 IncQuery Labs Ltd.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#  http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Created on 2019. nov. 12.

@author: Gabor Bergmann
'''

import collections
import importlib
from copy import copy

ApiClientModule = collections.namedtuple(
    "ApiClientModule", 
    ['endpoint_path', 
     'root_module_name',
     'api_names_to_class']
)

# TODO auto-generate from dir(iqs_client.api) + name mangling ? 
_api_client_modules = [
    ApiClientModule('api', 'iqs_client', {
        'acquisition'       : "AcquisitionApi",
        'async'             : "AsyncApi",
        'analysis'          : 'AnalysisApi',
        'impact_analysis'   : "ImpactAnalysisApi",
        'in_memory_index'   : "InMemoryIndexApi",
        'persistent_index'  : "PersistentIndexApi",
        'queries'           : "QueriesApi",
        'query_execution'   : "QueryExecutionApi",
        'repository'        : "RepositoryApi",
        'server_management' : "ServerManagementApi",
        'validation'        : "ValidationApi",    
        'integration'       : "IntegrationApi",
        'mms_repository'    : "MmsRepositoryApi",
        'experimental'      : "ExperimentalApi",
        'demo'              : "DemoApi"
    }),
]

def decorate_iqs_client(iqs_client_object, root_configuration):
    for api_client_module in _api_client_modules:
        try: # if client lib is missing, simply skip
            endpoint_root_module=importlib.import_module(api_client_module.root_module_name)
            endpoint_client_class = getattr(endpoint_root_module, "ApiClient")
            
            endpoint_specific_config = copy(root_configuration)
            endpoint_specific_config.host="{}/{}".format(root_configuration.host, api_client_module.endpoint_path)
            endpoint_specific_client = endpoint_client_class(endpoint_specific_config)

            for api_field_name, api_class_name in api_client_module.api_names_to_class.items():
                api_module_name= "{}.api.{}_api".format(api_client_module.root_module_name, api_field_name)
                try:
                    api_module=importlib.import_module(api_module_name)
                    if api_class_name in dir(api_module): # there might be api classes presently missing / disabled
                        api_class = getattr(api_module, api_class_name)
                        api_object = api_class(endpoint_specific_client)
                        setattr(iqs_client_object, api_field_name, api_object)
                except ImportError:
                    pass
        except ImportError:
            pass



# Copyright 2019 IncQuery Labs Ltd.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#  http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Created on 2019. máj. 24.

@author: Gabor Bergmann
'''

from typing import Optional, Callable, List, Any  # @UnusedImport

# connector-specific extensions may add tools such as browser widgets here
class IQSJupyterTools:
    def __init__(
        self,
        iqs
    ):
        self._iqs = iqs


# connector-specific extensions may add custom ways to convert a dict to an element representation; 
#     parse a dict and return the constructed element or None if not the right format
#     default is ElementInCompartmentDescriptor 
element_dict_recognizers : List[Callable[[dict], Optional[Any]]] = []




# Copyright 2019 IncQuery Labs Ltd.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#  http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Created on 2019. nov. 27.

@author: Gabor Bergmann
'''

import html

import iqs_jupyter.tool_extension_point as ext_point


def cell_to_html(cell):
    if hasattr(cell, '_repr_html_'):
        return cell._repr_html_()
    elif isinstance(cell, str):
        return html.escape(cell)
    else:
        return str(cell)

def dict_to_element(dict_or_attribute_value, url_provider=None):
    if isinstance(dict_or_attribute_value, dict):
        for recognizer in ext_point.element_dict_recognizers:
            recognized = recognizer(dict_or_attribute_value)
            if recognized:
                if url_provider is not None:
                    recognized.url = url_provider(recognized)
                return recognized
        # not recognized as an element, treated as raw dict    
        return dict_or_attribute_value
    else:
        return dict_or_attribute_value




# Copyright 2019 IncQuery Labs Ltd.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#  http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Created on 2019-05-17

@author: Gábor Bergmann
'''
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

# single-sourcing of version string specified in about.py
# see solution 3 @ https://packaging.python.org/guides/single-sourcing-package-version
version = {}
with open("iqs_jupyter/about.py") as fp:
    exec(fp.read(), version)

REQUIRES = ["incqueryserver-api-python-client=={}".format(version['__version__']), "mms-python-client", "ipywidgets"]

setuptools.setup(
    name="incqueryserver-jupyter",
    version=version['__version__'],
    author="Gábor Bergmann",
    author_email="gabor.bergmann@incquerylabs.com",
    description="IncQuery Server Client Extensions for Jupyter and Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/IncQueryLabs/incquery-server-jupyter",
    install_requires=REQUIRES,
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)
